/*

EEE ACPI Events Handling
Copyright (C) 2006  The State and University Library

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*
  eee_acpi 1.0

  The purpose of this program is to hangle ACPI events such as: 
  - Sound Fn-Keys : Vol Up, Vol Down and Mute;
  - Battery / AC adapter status;
  - External / Internal Display switch.
  
  It should be run at session startup by creating $HOME/.config/autostart/eee_acpi.desktop with this content: 
[Desktop Entry]
Name=eee_acpi
Comment=EEE ACPI user events handling
Exec=eee_acpi
Terminal=false
StartupNotify=false


  For more details, see http://lordikc.free.fr/wordpress

  eee_acpi requires libnotify.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <unistd.h>
#include <math.h>
#include <sys/time.h>

// Some constants
static const int MUTE    =1;
static const int VOLUP   =2;
static const int VOLDOWN =3;
static const int ACON    =4;
static const int ACOFF   =5;
static const int XRANDR  =6;
static const int BATT    =7;
static const int WIFION  =8;
static const int WIFIOFF =9;

static const int VGA_CONNECTED     =1;
static const int LVDS_CONNECTED    =2;
static const int VGA_ACTIVE        =4;
static const int LVDS_ACTIVE       =8;


// Variables and constants to avoid having too much messages
static double tpsLast=0;
static const int delaySound=700;
static const int delayNotif=3000;
char* lastNotif=NULL;

// TODO estimate battery lifetime.
static int battState = -1;
static time_t battTime = 0;

bool bBat = true;
bool bAc = true;
bool bVol = true;
bool bWifi = true;
bool bXr=true;
const char* extMode="";

// Read current event out of the acpid socket.
int read_event(int socket_fd)
{
  char msg[64];
  int index=0;
  bool found=false;
  while((index<64)&&!found)
    {
      char c;
      size_t res = read(socket_fd,&c,1);
      if(res<=0)
	return -1;
      else
	{
	  if(c=='\n')
	    {
	      // This is the end of the event
	      msg[index++]=0;
	      found=true;
	    }
	  else
	    {
	      // Accumulate chars
	      msg[index++]=c;
	    }
	}
    }
  if(found)
    {
      //      printf("%s\n",msg);
      if((strncmp(msg,"hotkey ATKD 00000013",20)==0)&&bVol) return MUTE;
      if((strncmp(msg,"hotkey ATKD 00000014",20)==0)&&bVol) return VOLDOWN;
      if((strncmp(msg,"hotkey ATKD 00000015",20)==0)&&bVol) return VOLUP;
      if((strncmp(msg,"hotkey ATKD 00000050",20)==0)&&bAc) return ACON;
      if((strncmp(msg,"hotkey ATKD 00000051",20)==0)&&bAc) return ACOFF;
      if((strncmp(msg,"hotkey ATKD 00000030",20)==0)&&bXr) return XRANDR;
      if((strncmp(msg,"hotkey ATKD 00000031",20)==0)&&bXr) return XRANDR;
      if((strncmp(msg,"hotkey ATKD 00000032",20)==0)&&bXr) return XRANDR;
      if((strncmp(msg,"battery BAT0 00000080",21)==0)&&bBat) return BATT;
      if((strncmp(msg,"hotkey ATKD 00000011",20)==0)&&bWifi) return WIFIOFF;
      if((strncmp(msg,"hotkey ATKD 00000010",20)==0)&&bWifi) return WIFION;
    }
  return -1;
}

// Execute a command and returns the output as a string
char* execute(const char * cmd)
{
  FILE* f = popen(cmd,"r");
  if(f!=NULL)
    {
      // Read at most 4096 chars. It should be sufficient
      // for the used command.
      char tmp[4096];
      size_t res = fread(tmp,1,4096,f);
      pclose(f);
      if(res>0)
	{
	  tmp[res]=0;
	  return strdup(tmp);
	}
    }
  return NULL;
}

// Notify the user through dbus and notification-daemon
// Requires the libnotify package providing notify-send.
void notify(const char* msg,int delayNotif)
{
  struct timeval tv;
  gettimeofday(&tv,NULL);
  double tps = (double)tv.tv_sec*1000.+(double)tv.tv_usec/1000.;
  // Ignoring notification if previous one is still displayed
  if(tps>tpsLast)
    {
      // Do not display the same notification twice
      if(strcmp(msg,lastNotif))
	{
	  free(lastNotif);
	  lastNotif = strdup(msg);
	  system(msg);
	}
      tpsLast=tps+delayNotif+delayNotif/10;
    }
}

// Notify the new volume level
// Notification look like
// Volume 10%
// [===|=======]
void notify_volume(const char* res)
{
  // Read volume out of the result from a amixer command
  if(res!=NULL)
    {
      // Search Mono:
      char* c = strstr(res,"Mono:");
      if(c!=NULL)
	{
	  // Retains the string between [45%]
	  c = strstr(c,"[");
	  if(c!=NULL)
	    {
	      c++;
	      char* c2=strstr(c,"]");
	      if (c2!=NULL)
		{
		  *c2=0;
		  char cmd[1024];
		  int vol=atoi(c);
		  sprintf(cmd,"notify-send -i sound_section -t %d \"Sound control\" \"Volume %s\n[",delaySound,c);
		  c=cmd+strlen(cmd);
		  vol=(int)round(.16*(double)vol);
		  // Construct an ascii volume bar looking like
		  // [====|======]
		  for(int i=0;i<vol;i++)
		    *c++='=';
		  *c++='|';
		  for(int i=vol;i<16;i++)
		    *c++='=';
		  
		  sprintf(c,"]\"");
		  
		  // Send notification to the user
		  notify(cmd,delaySound);
		}
	    }
	}
    }
}

// Switch internal/external display
void xrandr()
{
  char * res = execute("xrandr -q");
  int ret=0;
  bool bStartVGASection = false;
  bool bStartLVDSSection = false;
  for(char* c = strtok(res,"\n");c!=NULL;c = strtok(NULL,"\n"))
    {
      if(strstr(c,"VGA connected"))
	{
	  ret|=VGA_CONNECTED;
	  bStartVGASection=true;
	  bStartLVDSSection=false;
	}
      else if(strstr(c,"LVDS connected"))
	{
	  ret|=LVDS_CONNECTED;
	  bStartVGASection=false;
	  bStartLVDSSection=true;
	}
      else if (c[0]==' ')
	{
	  if(strstr(c,"*"))
	    {
	      if(bStartVGASection)
		ret|=VGA_ACTIVE;
	      if(bStartLVDSSection)
		ret|=LVDS_ACTIVE;
	    }
	}
      else
	{
	  bStartVGASection=false;
	  bStartLVDSSection=false;
	}
    }
  free(res);

  char cmd[1024];
  if(ret&VGA_CONNECTED)
    {
      // Both active, switch to VGA only
      if((ret&VGA_ACTIVE)&&(ret&LVDS_ACTIVE))
	{
	  if(strlen(extMode))
	    sprintf(cmd,"xrandr --output VGA --mode %s --output LVDS --off",extMode);
	  else
	    sprintf(cmd,"xrandr --output VGA --auto --output LVDS --off");
	  system(cmd);
	  sprintf(cmd,"notify-send -i configuration_section -t %d \"Display setting\" \"External display %s\"",delayNotif,strlen(extMode)?extMode:"");
	}
      // Only LVDS, switch to both active
      else if(!(ret&VGA_ACTIVE)&&(ret&LVDS_ACTIVE))
	{
	  system("xrandr --output VGA --mode 640x480 --output LVDS --mode 640x480");
	  sprintf(cmd,"notify-send -i configuration_section -t %d \"Display setting\" \"Both display\"",delayNotif);
	}
      // Only VGA switch to only LVDS
      else if((ret&VGA_ACTIVE)&&!(ret&LVDS_ACTIVE))
	{
	  system("xrandr --output VGA --off --output LVDS --auto");
	  sprintf(cmd,"notify-send -i configuration_section -t %d \"Display setting\" \"Internal display\"",delayNotif);
	}
    }
  else
    {
      // No external display connected restore defaults to LVDS
      system("xrandr --output VGA --off --output LVDS --auto");
      sprintf(cmd,"notify-send -i configuration_section -t %d \"Display setting\" \"Internal display\"",delayNotif);
    }
  notify(cmd,0);
}

int main (int argc, char *argv[])
{
  printf("  eee_acpi 1.0\n\
\n\
  The purpose of this program is to hangle ACPI events such as: \n\
  - Sound Fn-Keys : Vol Up, Vol Down and Mute;\n\
  - Battery / AC adapter status;\n\
  - External / Internal Display switch.\n\
  \n\
  It should be run at session startup.\n\
\n\
  eee_acpi requires libnotify.\n\
  For more details, see http://blogeee.free.fr/wordpress\n");

  for(int i=1;i<argc;i++)
    {
      if(!strcmp(argv[i],"-b")) bBat=false;
      if(!strcmp(argv[i],"-a")) bAc=false;
      if(!strcmp(argv[i],"-v")) bVol=false;
      if(!strcmp(argv[i],"-w")) bWifi=false;
      if(!strcmp(argv[i],"-x")) bXr=false;
      if(!strcmp(argv[i],"+x")) 
	{
	  bXr=true;
	  if(((i+1)<argc)&&strncmp(argv[i+1],"-",1))
	    {
	      i++;
	      extMode=argv[i];
	    }
	}
      if(!strcmp(argv[i],"-h"))
	{
	  printf("Usage : %s [options]\n",argv[0]);
	  printf("  -b : Disable batterie notifications\n");
	  printf("  -a : Disable AC notifications\n");
	  printf("  -v : Disable volume handling/notifications\n");
	  printf("  -x : Disable internal/external screen switch\n");
	  printf("  +x Mode : Enable internal/external scree switch with specifyied mode\n");
	  return 1;
	}
    }

  if(bBat) printf("Batterie notifications\n");
  if(bAc) printf("AC notifications\n");
  if(bVol) printf("Volume handling/notifications\n");
  if(bXr) printf("Internal/external screen switch\n");
  if(strlen(extMode)) printf("  Mode : %s\n",extMode);

  const char * acpidsock="/var/run/acpid.socket";
  int socket_fd = socket(AF_UNIX, SOCK_STREAM, 0);
  struct sockaddr_un socket_addr;
  socket_addr.sun_family = AF_UNIX;
  sprintf(socket_addr.sun_path, "%s", acpidsock);
  
  if(connect(socket_fd, (struct sockaddr *)&socket_addr, sizeof(socket_addr)) < 0)
    {
      printf("cannot connect socket\n");
      close(socket_fd);
      return -1;
    }

  lastNotif = strdup("");

  char tmp[1024];
  // Looping : read event, process event indifinetly.
  while (1)
    {
      char* res = NULL;
      int evt=read_event(socket_fd);
      switch(evt)
	{
	case MUTE:
	  res = execute("amixer -c 0 set Master toggle");
	  if(res!=NULL)		  
	    {
	      char cmd[1024];
	      sprintf(cmd,"notify-send -i sound_section -t %d \"Sound control\" \"%s\"",delayNotif,strstr(res,"[off]")?"Mute":"Unmute");
	      notify(cmd,0);
	    }
	  break;
	case VOLUP:
	  res = execute("amixer -c 0 set Master 6+");
	  notify_volume(res);
	  break;
	case VOLDOWN:
	  res = execute("amixer -c 0 set Master 6-");
	  notify_volume(res);
	  break;
	case ACON:
	case ACOFF:
	  {
	    char cmd[1024];
	    sprintf(cmd,"notify-send -i monitoring_section -t %d \"AC Adapter\" \"%s\"",delayNotif,evt==ACON?"Plugged":"Unplugged");
	    notify(cmd,0);
	  }
	  break;
	case XRANDR:
	  {
	    xrandr();
	  }
	  break;
	case BATT:
	  {
	    // Cut out "Battery 1: " and ", rate information unavailable."
	    res = execute("acpi -b");
	    if(res!=NULL)
	      {
		char* msg=strstr(res,":");
		if(msg!=NULL)
		  {
		    msg+=2;
		    char* end=strstr(res,"%");
		    if(end!=NULL)
		      {
			*(end+1)=0;
		      }
		    char cmd[1024];
		    sprintf(cmd,"notify-send -i monitoring_section -t %d \"Battery status\" \"%s\"",delayNotif,msg);
		    notify(cmd,0);
		  }
	      }
	  }
	  break;
	case WIFION:
	case WIFIOFF:
	  {
	    char cmd[1024];
	    sprintf(cmd,"notify-send -i networking_configuration_section -t %d \"WiFi\" \"%s\"",delayNotif,evt==WIFION?"Plugged":"Unplugged");
	    notify(cmd,0);
	  }
	  break;
	}
      if(res!=NULL)
	free(res);
    }
  close(socket_fd);
}
